"""
Clingo application class extended to solve COOM configuration problems
"""

import textwrap
from typing import Any, Callable, Dict, List, Optional, Sequence

from clingcon import ClingconTheory
from clingo import Control, Model, Symbol
from clingo.application import Application, ApplicationOptions, Flag
from clingo.ast import Location, Position, ProgramBuilder, Rule, parse_files
from clingo.symbol import Function, SymbolType
from flingo.__main__ import CSP, DEF, MAX_INT, MIN_INT
from flingo.__main__ import AppConfig as FlingoConfig
from flingo.__main__ import Statistic
from flingo.parsing import THEORY, HeadBodyTransformer
from flingo.translator import AUX, Translator

from .utils import asp2coom, get_encoding
from .utils.logging import get_logger

# mypy: allow-untyped-calls
log = get_logger("main")


def _get_valuation(model: Model) -> List[Symbol]:
    return [
        Function("value", assignment.arguments)
        for assignment in model.symbols(theory=True)
        if assignment.name == CSP
        and len(assignment.arguments) == 2
        and model.contains(
            Function(DEF, [Function(str(assignment.arguments[0]), [], True)])
        )  # Temporary fix until flingo fixes String behavior
        and not (assignment.arguments[0].type is SymbolType.Function and assignment.arguments[0].name == AUX)
    ]


def _sym_to_prg(symbols: Sequence[Symbol], output: Optional[str] = "asp") -> str:  # nocoverage
    """
    Turns symbols into a program.
    """
    sorted_symbols = sorted(symbols)
    if output == "asp":
        output_list = [f"{str(s)}" for s in sorted_symbols]
    elif output == "coom":
        output_list = [f"{asp2coom(s)}" for s in sorted_symbols]
    else:
        raise ValueError(f"Unrecognized output format: {output}")
    return "\n".join(output_list)


class COOMSolverApp(Application):
    """
    Application class extending clingo.
    """

    _options: Dict[str, Any]
    _istest: bool
    _log_level: str
    config: FlingoConfig
    _propagator: ClingconTheory
    program_name: str = "COOM Suite"
    version: str = "1.1.0"

    def __init__(
        self,
        log_level: str = "",
        options: Optional[Dict[str, Any]] = None,
        istest: bool = False,
    ):
        """
        Create application.
        """
        self._options = {"solver": "clingo", "output_format": "asp"} if options is None else options
        self._istest = istest
        self._log_level = "WARNING" if log_level == "" else log_level
        self.config = FlingoConfig(MIN_INT, MAX_INT, Flag(False), Flag(False), DEF)
        self._propagator = ClingconTheory()

    def parse_log_level(self, log_level: str) -> bool:  # nocoverage
        """
        Parse log.
        """
        if log_level is not None:
            self._log_level = log_level.upper()
            return self._log_level in ["INFO", "WARNING", "DEBUG", "ERROR"]

        return True

    def register_options(self, options: ApplicationOptions) -> None:  # nocoverage
        """
        Add custom options.
        """
        group = "Clingo.COOM"
        options.add(
            group,
            "log",
            textwrap.dedent(
                """\
                Provide logging level.
                                            <level> ={debug|info|error|warning}
                                            (default: warning)"""
            ),
            self.parse_log_level,
            argument="<level>",
        )
        # options.add_flag(
        #     group, "view", "Visualize the first solution using clinguin", self._view
        # )

    def on_model(self, model: Model) -> None:  # nocoverage
        """
        Function called after finding each model.
        Args:
            model (Model): clingo Model
        """
        if self._options["solver"] == "flingo":
            self._propagator.on_model(model)

            if self._istest:
                # this slows down solving considerably but makes tests for clingo and flingo uniform
                # better way?
                model.extend(_get_valuation(model))
        # log.debug("------- Full model -----") # disabled because makes solving much slower
        # log.debug("\n".join([str(s) for s in model.symbols(atoms=True, shown=True, theory=True)]))

    def print_model(self, model: Model, printer: Callable[[], None]) -> None:  # nocoverage
        """
        Print a model on the console.
        """

        if self._options["solver"] == "clingo":
            output_symbols = model.symbols(shown=True)
        elif self._options["solver"] == "flingo":
            output_symbols = [
                atom
                for atom in model.symbols(shown=True)
                if not (atom.name == self.config.defined and len(atom.arguments) == 1)
            ]
            output_symbols.extend(_get_valuation(model))
        print(_sym_to_prg(output_symbols, self._options["output_format"]))

    def main(self, control: Control, files: Sequence[str]) -> None:
        """
        Main function ran on call.
        """
        encoding = get_encoding(f"encoding-base-{self._options['solver']}.lp")
        show = get_encoding(f"show-{self._options['solver']}.lp")
        for f in files:
            control.load(f)

        if self._options["solver"] == "clingo":
            control.load(encoding)
            control.load(show)
            control.ground()
            control.solve()

        elif self._options["solver"] == "flingo":
            self._propagator.register(control)
            self._propagator.configure("max-int", str(self.config.max_int))
            self._propagator.configure("min-int", str(self.config.min_int))

            control.add("base", [], THEORY)

            with ProgramBuilder(control) as bld:
                hbt = HeadBodyTransformer()

                parse_files([encoding, show], lambda ast: bld.add(hbt.visit(ast)))
                pos = Position("<string>", 1, 1)
                loc = Location(pos, pos)
                for rule in hbt.rules_to_add:
                    bld.add(Rule(loc, rule[0], rule[1]))  # nocoverage # Not sure when this is needed

            control.ground([("base", [])])
            translator = Translator(control, self.config, Statistic())
            translator.translate(control.theory_atoms)

            self._propagator.prepare(control)
            control.solve(on_model=self.on_model, on_statistics=self._propagator.on_statistics)



# Generated from /home/heuer/tools/coom-suite/src/coomsuite/utils/coom_grammar/model/Model.g4 by ANTLR 4.13.2
from antlr4 import *
if "." in __name__:
    from .ModelParser import ModelParser
else:
    from ModelParser import ModelParser

# This class defines a complete listener for a parse tree produced by ModelParser.
class ModelListener(ParseTreeListener):

    # Enter a parse tree produced by ModelParser#root.
    def enterRoot(self, ctx:ModelParser.RootContext):
        pass

    # Exit a parse tree produced by ModelParser#root.
    def exitRoot(self, ctx:ModelParser.RootContext):
        pass


    # Enter a parse tree produced by ModelParser#product.
    def enterProduct(self, ctx:ModelParser.ProductContext):
        pass

    # Exit a parse tree produced by ModelParser#product.
    def exitProduct(self, ctx:ModelParser.ProductContext):
        pass


    # Enter a parse tree produced by ModelParser#structure.
    def enterStructure(self, ctx:ModelParser.StructureContext):
        pass

    # Exit a parse tree produced by ModelParser#structure.
    def exitStructure(self, ctx:ModelParser.StructureContext):
        pass


    # Enter a parse tree produced by ModelParser#enumeration.
    def enterEnumeration(self, ctx:ModelParser.EnumerationContext):
        pass

    # Exit a parse tree produced by ModelParser#enumeration.
    def exitEnumeration(self, ctx:ModelParser.EnumerationContext):
        pass


    # Enter a parse tree produced by ModelParser#feature.
    def enterFeature(self, ctx:ModelParser.FeatureContext):
        pass

    # Exit a parse tree produced by ModelParser#feature.
    def exitFeature(self, ctx:ModelParser.FeatureContext):
        pass


    # Enter a parse tree produced by ModelParser#cardinality.
    def enterCardinality(self, ctx:ModelParser.CardinalityContext):
        pass

    # Exit a parse tree produced by ModelParser#cardinality.
    def exitCardinality(self, ctx:ModelParser.CardinalityContext):
        pass


    # Enter a parse tree produced by ModelParser#attribute.
    def enterAttribute(self, ctx:ModelParser.AttributeContext):
        pass

    # Exit a parse tree produced by ModelParser#attribute.
    def exitAttribute(self, ctx:ModelParser.AttributeContext):
        pass


    # Enter a parse tree produced by ModelParser#option.
    def enterOption(self, ctx:ModelParser.OptionContext):
        pass

    # Exit a parse tree produced by ModelParser#option.
    def exitOption(self, ctx:ModelParser.OptionContext):
        pass


    # Enter a parse tree produced by ModelParser#field.
    def enterField(self, ctx:ModelParser.FieldContext):
        pass

    # Exit a parse tree produced by ModelParser#field.
    def exitField(self, ctx:ModelParser.FieldContext):
        pass


    # Enter a parse tree produced by ModelParser#number_def.
    def enterNumber_def(self, ctx:ModelParser.Number_defContext):
        pass

    # Exit a parse tree produced by ModelParser#number_def.
    def exitNumber_def(self, ctx:ModelParser.Number_defContext):
        pass


    # Enter a parse tree produced by ModelParser#string_def.
    def enterString_def(self, ctx:ModelParser.String_defContext):
        pass

    # Exit a parse tree produced by ModelParser#string_def.
    def exitString_def(self, ctx:ModelParser.String_defContext):
        pass


    # Enter a parse tree produced by ModelParser#unit.
    def enterUnit(self, ctx:ModelParser.UnitContext):
        pass

    # Exit a parse tree produced by ModelParser#unit.
    def exitUnit(self, ctx:ModelParser.UnitContext):
        pass


    # Enter a parse tree produced by ModelParser#fraction.
    def enterFraction(self, ctx:ModelParser.FractionContext):
        pass

    # Exit a parse tree produced by ModelParser#fraction.
    def exitFraction(self, ctx:ModelParser.FractionContext):
        pass


    # Enter a parse tree produced by ModelParser#behavior.
    def enterBehavior(self, ctx:ModelParser.BehaviorContext):
        pass

    # Exit a parse tree produced by ModelParser#behavior.
    def exitBehavior(self, ctx:ModelParser.BehaviorContext):
        pass


    # Enter a parse tree produced by ModelParser#behavior_block.
    def enterBehavior_block(self, ctx:ModelParser.Behavior_blockContext):
        pass

    # Exit a parse tree produced by ModelParser#behavior_block.
    def exitBehavior_block(self, ctx:ModelParser.Behavior_blockContext):
        pass


    # Enter a parse tree produced by ModelParser#define.
    def enterDefine(self, ctx:ModelParser.DefineContext):
        pass

    # Exit a parse tree produced by ModelParser#define.
    def exitDefine(self, ctx:ModelParser.DefineContext):
        pass


    # Enter a parse tree produced by ModelParser#conditioned.
    def enterConditioned(self, ctx:ModelParser.ConditionedContext):
        pass

    # Exit a parse tree produced by ModelParser#conditioned.
    def exitConditioned(self, ctx:ModelParser.ConditionedContext):
        pass


    # Enter a parse tree produced by ModelParser#exists.
    def enterExists(self, ctx:ModelParser.ExistsContext):
        pass

    # Exit a parse tree produced by ModelParser#exists.
    def exitExists(self, ctx:ModelParser.ExistsContext):
        pass


    # Enter a parse tree produced by ModelParser#explanation.
    def enterExplanation(self, ctx:ModelParser.ExplanationContext):
        pass

    # Exit a parse tree produced by ModelParser#explanation.
    def exitExplanation(self, ctx:ModelParser.ExplanationContext):
        pass


    # Enter a parse tree produced by ModelParser#precondition.
    def enterPrecondition(self, ctx:ModelParser.PreconditionContext):
        pass

    # Exit a parse tree produced by ModelParser#precondition.
    def exitPrecondition(self, ctx:ModelParser.PreconditionContext):
        pass


    # Enter a parse tree produced by ModelParser#date.
    def enterDate(self, ctx:ModelParser.DateContext):
        pass

    # Exit a parse tree produced by ModelParser#date.
    def exitDate(self, ctx:ModelParser.DateContext):
        pass


    # Enter a parse tree produced by ModelParser#combinations.
    def enterCombinations(self, ctx:ModelParser.CombinationsContext):
        pass

    # Exit a parse tree produced by ModelParser#combinations.
    def exitCombinations(self, ctx:ModelParser.CombinationsContext):
        pass


    # Enter a parse tree produced by ModelParser#combination_row.
    def enterCombination_row(self, ctx:ModelParser.Combination_rowContext):
        pass

    # Exit a parse tree produced by ModelParser#combination_row.
    def exitCombination_row(self, ctx:ModelParser.Combination_rowContext):
        pass


    # Enter a parse tree produced by ModelParser#combination_item.
    def enterCombination_item(self, ctx:ModelParser.Combination_itemContext):
        pass

    # Exit a parse tree produced by ModelParser#combination_item.
    def exitCombination_item(self, ctx:ModelParser.Combination_itemContext):
        pass


    # Enter a parse tree produced by ModelParser#combination_atom.
    def enterCombination_atom(self, ctx:ModelParser.Combination_atomContext):
        pass

    # Exit a parse tree produced by ModelParser#combination_atom.
    def exitCombination_atom(self, ctx:ModelParser.Combination_atomContext):
        pass


    # Enter a parse tree produced by ModelParser#assign_default.
    def enterAssign_default(self, ctx:ModelParser.Assign_defaultContext):
        pass

    # Exit a parse tree produced by ModelParser#assign_default.
    def exitAssign_default(self, ctx:ModelParser.Assign_defaultContext):
        pass


    # Enter a parse tree produced by ModelParser#assign_imply.
    def enterAssign_imply(self, ctx:ModelParser.Assign_implyContext):
        pass

    # Exit a parse tree produced by ModelParser#assign_imply.
    def exitAssign_imply(self, ctx:ModelParser.Assign_implyContext):
        pass


    # Enter a parse tree produced by ModelParser#optimize.
    def enterOptimize(self, ctx:ModelParser.OptimizeContext):
        pass

    # Exit a parse tree produced by ModelParser#optimize.
    def exitOptimize(self, ctx:ModelParser.OptimizeContext):
        pass


    # Enter a parse tree produced by ModelParser#interaction.
    def enterInteraction(self, ctx:ModelParser.InteractionContext):
        pass

    # Exit a parse tree produced by ModelParser#interaction.
    def exitInteraction(self, ctx:ModelParser.InteractionContext):
        pass


    # Enter a parse tree produced by ModelParser#message.
    def enterMessage(self, ctx:ModelParser.MessageContext):
        pass

    # Exit a parse tree produced by ModelParser#message.
    def exitMessage(self, ctx:ModelParser.MessageContext):
        pass


    # Enter a parse tree produced by ModelParser#require.
    def enterRequire(self, ctx:ModelParser.RequireContext):
        pass

    # Exit a parse tree produced by ModelParser#require.
    def exitRequire(self, ctx:ModelParser.RequireContext):
        pass


    # Enter a parse tree produced by ModelParser#prefer.
    def enterPrefer(self, ctx:ModelParser.PreferContext):
        pass

    # Exit a parse tree produced by ModelParser#prefer.
    def exitPrefer(self, ctx:ModelParser.PreferContext):
        pass


    # Enter a parse tree produced by ModelParser#condition.
    def enterCondition(self, ctx:ModelParser.ConditionContext):
        pass

    # Exit a parse tree produced by ModelParser#condition.
    def exitCondition(self, ctx:ModelParser.ConditionContext):
        pass


    # Enter a parse tree produced by ModelParser#condition_or.
    def enterCondition_or(self, ctx:ModelParser.Condition_orContext):
        pass

    # Exit a parse tree produced by ModelParser#condition_or.
    def exitCondition_or(self, ctx:ModelParser.Condition_orContext):
        pass


    # Enter a parse tree produced by ModelParser#condition_and.
    def enterCondition_and(self, ctx:ModelParser.Condition_andContext):
        pass

    # Exit a parse tree produced by ModelParser#condition_and.
    def exitCondition_and(self, ctx:ModelParser.Condition_andContext):
        pass


    # Enter a parse tree produced by ModelParser#condition_not.
    def enterCondition_not(self, ctx:ModelParser.Condition_notContext):
        pass

    # Exit a parse tree produced by ModelParser#condition_not.
    def exitCondition_not(self, ctx:ModelParser.Condition_notContext):
        pass


    # Enter a parse tree produced by ModelParser#condition_compare.
    def enterCondition_compare(self, ctx:ModelParser.Condition_compareContext):
        pass

    # Exit a parse tree produced by ModelParser#condition_compare.
    def exitCondition_compare(self, ctx:ModelParser.Condition_compareContext):
        pass


    # Enter a parse tree produced by ModelParser#condition_part.
    def enterCondition_part(self, ctx:ModelParser.Condition_partContext):
        pass

    # Exit a parse tree produced by ModelParser#condition_part.
    def exitCondition_part(self, ctx:ModelParser.Condition_partContext):
        pass


    # Enter a parse tree produced by ModelParser#formula.
    def enterFormula(self, ctx:ModelParser.FormulaContext):
        pass

    # Exit a parse tree produced by ModelParser#formula.
    def exitFormula(self, ctx:ModelParser.FormulaContext):
        pass


    # Enter a parse tree produced by ModelParser#formula_add.
    def enterFormula_add(self, ctx:ModelParser.Formula_addContext):
        pass

    # Exit a parse tree produced by ModelParser#formula_add.
    def exitFormula_add(self, ctx:ModelParser.Formula_addContext):
        pass


    # Enter a parse tree produced by ModelParser#formula_sub.
    def enterFormula_sub(self, ctx:ModelParser.Formula_subContext):
        pass

    # Exit a parse tree produced by ModelParser#formula_sub.
    def exitFormula_sub(self, ctx:ModelParser.Formula_subContext):
        pass


    # Enter a parse tree produced by ModelParser#formula_mul.
    def enterFormula_mul(self, ctx:ModelParser.Formula_mulContext):
        pass

    # Exit a parse tree produced by ModelParser#formula_mul.
    def exitFormula_mul(self, ctx:ModelParser.Formula_mulContext):
        pass


    # Enter a parse tree produced by ModelParser#formula_div.
    def enterFormula_div(self, ctx:ModelParser.Formula_divContext):
        pass

    # Exit a parse tree produced by ModelParser#formula_div.
    def exitFormula_div(self, ctx:ModelParser.Formula_divContext):
        pass


    # Enter a parse tree produced by ModelParser#formula_pow.
    def enterFormula_pow(self, ctx:ModelParser.Formula_powContext):
        pass

    # Exit a parse tree produced by ModelParser#formula_pow.
    def exitFormula_pow(self, ctx:ModelParser.Formula_powContext):
        pass


    # Enter a parse tree produced by ModelParser#formula_sign.
    def enterFormula_sign(self, ctx:ModelParser.Formula_signContext):
        pass

    # Exit a parse tree produced by ModelParser#formula_sign.
    def exitFormula_sign(self, ctx:ModelParser.Formula_signContext):
        pass


    # Enter a parse tree produced by ModelParser#formula_func.
    def enterFormula_func(self, ctx:ModelParser.Formula_funcContext):
        pass

    # Exit a parse tree produced by ModelParser#formula_func.
    def exitFormula_func(self, ctx:ModelParser.Formula_funcContext):
        pass


    # Enter a parse tree produced by ModelParser#formula_atom.
    def enterFormula_atom(self, ctx:ModelParser.Formula_atomContext):
        pass

    # Exit a parse tree produced by ModelParser#formula_atom.
    def exitFormula_atom(self, ctx:ModelParser.Formula_atomContext):
        pass


    # Enter a parse tree produced by ModelParser#constant.
    def enterConstant(self, ctx:ModelParser.ConstantContext):
        pass

    # Exit a parse tree produced by ModelParser#constant.
    def exitConstant(self, ctx:ModelParser.ConstantContext):
        pass


    # Enter a parse tree produced by ModelParser#floating.
    def enterFloating(self, ctx:ModelParser.FloatingContext):
        pass

    # Exit a parse tree produced by ModelParser#floating.
    def exitFloating(self, ctx:ModelParser.FloatingContext):
        pass


    # Enter a parse tree produced by ModelParser#path.
    def enterPath(self, ctx:ModelParser.PathContext):
        pass

    # Exit a parse tree produced by ModelParser#path.
    def exitPath(self, ctx:ModelParser.PathContext):
        pass


    # Enter a parse tree produced by ModelParser#path_item.
    def enterPath_item(self, ctx:ModelParser.Path_itemContext):
        pass

    # Exit a parse tree produced by ModelParser#path_item.
    def exitPath_item(self, ctx:ModelParser.Path_itemContext):
        pass


    # Enter a parse tree produced by ModelParser#path_index.
    def enterPath_index(self, ctx:ModelParser.Path_indexContext):
        pass

    # Exit a parse tree produced by ModelParser#path_index.
    def exitPath_index(self, ctx:ModelParser.Path_indexContext):
        pass


    # Enter a parse tree produced by ModelParser#name.
    def enterName(self, ctx:ModelParser.NameContext):
        pass

    # Exit a parse tree produced by ModelParser#name.
    def exitName(self, ctx:ModelParser.NameContext):
        pass


    # Enter a parse tree produced by ModelParser#stmt_end.
    def enterStmt_end(self, ctx:ModelParser.Stmt_endContext):
        pass

    # Exit a parse tree produced by ModelParser#stmt_end.
    def exitStmt_end(self, ctx:ModelParser.Stmt_endContext):
        pass


    # Enter a parse tree produced by ModelParser#compare.
    def enterCompare(self, ctx:ModelParser.CompareContext):
        pass

    # Exit a parse tree produced by ModelParser#compare.
    def exitCompare(self, ctx:ModelParser.CompareContext):
        pass



del ModelParser



"""
Test cases for checking the user input for validity.
"""

from os.path import join
from unittest import TestCase

from coomsuite.preprocess import check_user_input
from coomsuite.utils.logging import get_logger

from .clintests.tests_solve import TESTS_SOLVE

log = get_logger("main")


class TestUserInputCheck(TestCase):
    """
    Test user input check
    """

    def user_check(self, test: str, expected_msg: str) -> None:
        """
        Runs a test checking the user input for validity.
        """
        test_dict = TESTS_SOLVE[test]
        program = test_dict.get("program", None)
        if not program:
            program = []
            with open(join("examples", "tests", "solve", f"{test}.lp"), encoding="utf-8") as f:
                program.extend(f.readlines())
        with self.assertLogs(log, level="WARNING") as ctx:
            check_user_input(program)
        self.assertEqual(ctx.output, [f"WARNING:main:{expected_msg}"])

    def test_user_input_check(self) -> None:
        """
        Test user input check
        """
        self.user_check("set_invalid_variable", 'Invalid user input.\nVariable "root.color[0]" does not exist.')
        self.user_check("add_invalid_variable", 'Invalid user input.\nVariable "root.basket[0]" does not exist.')
        self.user_check(
            "set_invalid_type",
            'Invalid user input.\nNo value can be set for variable "root.basket[0]". Variable exists but is a part.',
        )
        # self.user_check(
        #     "add_invalid_type", "Invalid user input.\nVariable root.basket[0] cannot be added: Not a part."
        # )
        self.user_check(
            "set_invalid_value_discrete",
            'Invalid user input.\nValue "Yellow" is not in domain of variable "root.color[0]".',
        )
        self.user_check(
            "set_invalid_value_num", 'Invalid user input.\nValue "11" is not in domain of variable "root.size[0]".'
        )



"""
Test cases for multishot application class.
"""

# pylint: disable=protected-access

from contextlib import redirect_stdout
from typing import Any, List, Set, Tuple
from unittest import TestCase
from unittest.mock import ANY, call, create_autospec, patch

from clingo import Control, SolveResult
from clingo.symbol import Function, Number, String, Symbol, parse_term

from coomsuite.bounds.multi_application import COOMMultiSolverApp, _get_fact_name_and_args


class TestMultiApplication(TestCase):
    """
    Test cases for multishot application class.
    """

    def test_multi_application_helper_function(self) -> None:
        """
        Test helper function used in multishot application class.
        """
        for fact, result in [
            (
                'set("root.totalVolume","root.totalVolume[0]").',
                ("set", (parse_term('"root.totalVolume"'), parse_term('"root.totalVolume[0]"'))),
            ),
            (
                'index("root.color[0]",0).',
                ("index", (parse_term('"root.color[0]"'), parse_term("0"))),
            ),
            ('allow(7,(1,0),"S").', ("allow", (parse_term("7"), parse_term("(1,0)"), parse_term('"S"')))),
        ]:
            self.assertEqual(_get_fact_name_and_args(fact), result, f"failed with fact={fact}, result={result}")

        # test behavior when fact is not given with trailing "."
        self.assertRaises(RuntimeError, _get_fact_name_and_args, "p(0)")

    def test_init_flingo(self) -> None:
        """
        Test initialization with flingo as solver.
        """
        self.assertRaises(ValueError, COOMMultiSolverApp, [], options={"solver": "flingo"})

    def test_get_prog_part(self) -> None:
        """
        Test function for getting the program part of a (non-incremental) fact.
        """
        app = COOMMultiSolverApp([])

        for fact, part in [
            ('allow(7,(1,0),"S").', ("new_allow", (parse_term("7"), parse_term("(1,0)"), parse_term('"S"')))),
            ('number("5",5).', ("new_number", (parse_term('"5"'), parse_term("5")))),
            # program parts that get the bound added to arguments
            (
                'type("root.bags[1]","Bags").',
                ("new_type", (parse_term('"root.bags[1]"'), parse_term('"Bags"'), parse_term("1"))),
            ),
            (
                'constraint(("root.color",1),"lowerbound").',
                ("new_constraint", (parse_term('("root.color",1)'), parse_term('"lowerbound"'), parse_term("1"))),
            ),
            (
                'column(("Size","root.bags[0].size[0]"),0,0,"root.bags[0].size[0]").',
                (
                    "new_column",
                    (
                        parse_term('("Size","root.bags[0].size[0]")'),
                        parse_term("0"),
                        parse_term("0"),
                        parse_term('"root.bags[0].size[0]"'),
                        parse_term("1"),
                    ),
                ),
            ),
        ]:
            self.assertEqual(app._get_prog_part(fact, 1), part, f"failed with fact={fact}, part={part}")

    def test_get_incremental_prog_part(self) -> None:
        """
        Test function for getting the program part of an incremental fact.
        """
        app = COOMMultiSolverApp([])

        name = parse_term('"name"')
        bound = 3
        bound_term = Number(3)

        # test unary, constraint and function
        for exp_type, args, part in [
            ("unary", (name,), ("incremental_unary", (name, bound_term))),
            ("constraint", (name,), ("incremental_constraint", (name, bound_term))),
            # first the function prog parts has prefix new
            ("function", (name,), ("new_incremental_function", (name, bound_term))),
            # then it has prefix update
            ("function", (name,), ("update_incremental_function", (name, bound_term))),
        ]:
            self.assertEqual(
                app._get_incremental_prog_part(exp_type, args, bound),
                part,
                f"failed with exp_type={exp_type}, args={args}, part={part}",
            )

        # test three cases for binary expression
        lhs_name = parse_term('"lhs"')
        op_name = parse_term('"op"')
        rhs_name = parse_term('"rhs"')
        binary_args = (name, lhs_name, op_name, rhs_name)
        part_args = binary_args + (bound_term,)
        for inc_expressions, part_name in [
            # 1: only the right sub-expression is incremental
            ({rhs_name.string}, "incremental_binary_r"),
            # 2: only the left sub-expression is incremental
            ({lhs_name.string}, "incremental_binary_l"),
            # 3: both sub-expressions are incremental
            ({rhs_name.string, lhs_name.string}, "incremental_binary"),
        ]:
            app._incremental_expressions = inc_expressions
            self.assertEqual(
                app._get_incremental_prog_part("binary", binary_args, bound),
                (part_name, part_args),
                f"failed with inc_expressions={inc_expressions}, part_name={part_name}",
            )

        # test invalid incremental fact
        self.assertRaises(ValueError, app._get_incremental_prog_part, "number", [], 0)

    def test_update_bound(self) -> None:
        """
        Test the update bounds function of multi application class.
        """
        app = COOMMultiSolverApp([], algorithm="linear", initial_bound=3)

        for prev, current in [(None, 3), (3, 4), (4, 5)]:
            fail_msg = f"failed with prev={prev}, current={current}"
            self.assertEqual(app._prev_bound, prev, fail_msg)
            self.assertEqual(app.current_max_bound, current, fail_msg)
            app._update_bound()

        app = COOMMultiSolverApp([], algorithm="exponential", initial_bound=3)

        for prev, current in [(None, 3), (3, 4), (4, 8)]:
            fail_msg = f"failed with prev={prev}, current={current}"
            self.assertEqual(app._prev_bound, prev, fail_msg)
            self.assertEqual(app.current_max_bound, current, fail_msg)
            app._update_bound()

    def test_remove_new_incremental_expressions(self) -> None:
        """
        Test functionality of removing new incremental expressions from a list of facts.
        """
        app = COOMMultiSolverApp([])

        # initial value of _processed_facts
        processed = {'allow(7,(0,0),"small").'}

        # initial value of _new_processed_facts
        new = {
            'domain("Size","small").',
            'constraint(("root.color",1),"lowerbound").',
            'constraint((7,"root.bags[0]"),"table").',
            'constraint((2,"root.color[0]=Blue"),"boolean").',
            'function("count(root.bags[0].pockets)","count","root.bags[0].pockets").',
            'binary("root.color[0]=Blue","root.color[0]","=","Blue").',
            'unary("-7","-","7").',
        }
        # incremental expressions part of the initial value of _new_processed_facts
        incremental = {
            'constraint((4,"5<count(root.bags.pockets)"),"boolean").',
            'function("count(root.bags.pockets)","count","root.bags.pockets").',
            'binary("5<count(root.bags.pockets)","5","<","count(root.bags.pockets)").',
            'unary("(count(root.bags.pockets))","()","count(root.bags.pockets)").',
        }

        # initialize attributes accordingly
        app._new_processed_facts = new | incremental
        app._processed_facts = processed.copy()
        app._incremental_expressions = {
            "5<count(root.bags.pockets)",
            "count(root.bags.pockets)",
            "(count(root.bags.pockets))",
            "modules",
        }

        # remove the new incremental expressions
        removed = app._remove_new_incremental_expressions()

        # check that return value of function matches the incremental expressions
        # note that return value has type List[Tuple[str, List[Symbol]]]
        self.assertCountEqual(removed, [_get_fact_name_and_args(f) for f in incremental])
        # check that _processed_facts are the initial value (processed) + incremental
        self.assertEqual(app._processed_facts, processed | incremental)
        # check that _new_processed_facts contains only new but not incremental
        self.assertEqual(app._new_processed_facts, new)

    def test_update_incremental_data(self) -> None:
        """
        Test updating of data structures tracking incremental sets and expressions.
        """
        app = COOMMultiSolverApp([])

        # initial incremental data
        inc_parts: Set[Tuple[str, Tuple[Symbol, ...]]] = set()
        inc_parts = {
            (
                "function",
                (String("count(root.bags.pockets)"), String("count"), String("root.bags.pockets")),
            ),
        }
        inc_expressions = {"5<count(root.bags.pockets)"}

        # initialize attributes
        app._incremental_parts = inc_parts.copy()
        app._incremental_expressions = inc_expressions.copy()

        # incremental facts for updating incremental data
        incremental_facts = {
            'inc_set("root.bags.size.volume").',
            'inc_set("root.bags.pockets").',
            (
                'incremental("function","sum(root.bags.size.volume)",'
                '("sum(root.bags.size.volume)","sum","root.bags.size.volume")).'
            ),
            (
                'incremental("binary","5<count(root.bags.pockets)",'
                '("5<count(root.bags.pockets)","5","<","count(root.bags.pockets)")).'
            ),
            'incremental("constraint","5<count(root.bags.pockets)",((4,"5<count(root.bags.pockets)"),"boolean")).',
        }

        app._update_incremental_data(incremental_facts)

        # add everything to the expected results
        inc_parts.add(
            (
                "constraint",
                (parse_term('(4,"5<count(root.bags.pockets)")'), String("boolean")),
            )
        )
        inc_parts.add(
            (
                "binary",
                (String("5<count(root.bags.pockets)"), String("5"), String("<"), String("count(root.bags.pockets)")),
            )
        )
        inc_parts.add(
            (
                "function",
                (String("sum(root.bags.size.volume)"), String("sum"), String("root.bags.size.volume")),
            )
        )
        inc_expressions.add("sum(root.bags.size.volume)")

        # test value of _incremental_set dictionary
        self.assertEqual(app._incremental_parts, inc_parts)
        # test value of _incremental_expressions set
        self.assertEqual(app._incremental_expressions, inc_expressions)

    def _get_mock_control(self, solve_is_sat: List[bool]) -> Control:
        """
        Helper function to create a mock control object with a list of return values of solve.

        Args:
            solve_is_sat (List[bool]): list of returns values for solve calls (True/False for SAT/UNSAT)
        """
        control = create_autospec(Control)

        returns = []
        for sat in solve_is_sat:
            ret = create_autospec(SolveResult)
            ret.satisfiable = sat
            returns.append(ret)

        control.solve.side_effect = returns

        return control  # type: ignore[no-any-return]

    def test_find_minimal_bound(self) -> None:
        """
        Test the function computing the minimal bound.
        """
        app = COOMMultiSolverApp([])

        # expected function calls to the control object
        # - helper functions to define sequence of calls
        def assign(name: str, bound: int, value: bool) -> Any:
            return call.assign_external(Function(name, [Number(bound)], True), value)

        def release(name: str, bound: int) -> Any:
            return call.release_external(Function(name, [Number(bound)], True))

        # - initial calls
        initial_calls = [
            assign("active", 7, False),
            assign("active", 8, False),
            release("max_bound", 8),
            assign("max_bound", 6, True),
            call.solve(),
        ]
        # - calls after the first solve is satisfiable
        calls_after_sat = [
            assign("active", 6, False),
            release("max_bound", 6),
            assign("max_bound", 5, True),
            call.solve(),
        ]
        # - calls after the first solve is unsatisfiable
        calls_after_unsat = [
            assign("active", 7, True),
            release("max_bound", 6),
            assign("max_bound", 7, True),
            call.solve(),
        ]

        # define values for testing
        # prev_init: initial value of app._prev_bound
        # max_init: initial value of app.max_bound
        # sat_values: list of return values of control.solve for defining mock control
        # calls: expected sequence of calls to the control object
        # max_return: the expected value of app.max_bound upon return
        for prev_init, max_init, sat_values, calls, max_return in [
            (
                0,
                1,
                [],
                [],
                1,
            ),
            (
                4,
                8,
                [False, True],
                initial_calls + calls_after_unsat,
                7,
            ),
            (
                4,
                8,
                [False, False],
                initial_calls + calls_after_unsat,
                8,
            ),
            (
                4,
                8,
                [True, True],
                initial_calls + calls_after_sat,
                5,
            ),
            (
                4,
                8,
                [True, False],
                initial_calls + calls_after_sat,
                6,
            ),
        ]:
            # initialize values
            app._prev_bound = prev_init
            app.current_max_bound = max_init
            # initialize mock control object
            control = self._get_mock_control(sat_values)

            # compute the minimal bound
            with redirect_stdout(None):
                app._find_minimal_bound(control)

            fail_msg = (
                f"failed with prev_init={prev_init}, max_init={max_init}, "
                f"sat_values={sat_values}, max_return={max_return}"
            )

            # check that the correct minimal bound was computed
            self.assertEqual(app.current_max_bound, max_return, fail_msg)

            # check what calls were made to the control object
            self.assertEqual(control.mock_calls, calls, fail_msg)  # type: ignore[attr-defined]

    def test_preprocess_new_bound(self) -> None:
        """
        Test preprocessing and updates to data structures for a new bound.
        """
        # initial values for preocessed and new processed facts
        initial_processed = {'domain("Color","Green").'}
        initial_new_processed = {'domain("Color","Blue").'}

        # new facts added by preprocessing (mocked below)
        non_incremental = {'domain("Color","Red").'}
        incremental = {
            'inc_set("root.bags.pockets").',
            (
                'incremental("function","count(root.bags.pockets)","root.bags.pockets",'
                '("count(root.bags.pockets)","count","root.bags.pockets"))'
            ),
        }

        # define expected values after calling preprocess_new_bound
        # expected processed facts
        expected_processed = initial_processed | initial_new_processed
        # expected new processed facts
        expected_new_processed = non_incremental
        # expected arguments to update incremental data method
        expected_update_args = incremental

        # bound to process at
        bound = 2

        # mock preprocess function
        with patch("coomsuite.bounds.multi_application.preprocess", autospec=True) as mock_preprocess:
            # mock return value
            mock_preprocess.side_effect = [list(non_incremental | incremental)]

            # initialize multi solver
            app = COOMMultiSolverApp([])
            app._processed_facts = initial_processed.copy()
            app._new_processed_facts = initial_new_processed.copy()

            # mock update incremental data function
            with patch.object(app, "_update_incremental_data", autospec=True) as mock_update:
                # call multi solver preprocess function
                app._preprocess_new_bound(bound)

                # check arguments of call to preprocess
                self.assertEqual(mock_preprocess.call_args, call([], max_bound=bound, discrete=True, multishot=True))
                # check _processed_facts
                self.assertEqual(app._processed_facts, expected_processed)
                # check _new_processed_facts
                self.assertEqual(app._new_processed_facts, expected_new_processed)
                # check arguments of call to _update_incremental_data
                self.assertEqual(mock_update.call_args, call(expected_update_args))

    def test_compute_prog_parts_at_zero(self) -> None:
        """
        Test compute program parts function for bound zero.
        """
        app = COOMMultiSolverApp([])

        with (
            patch.object(app, "_remove_new_incremental_expressions", autospec=True) as mock_remove,
            patch.object(app, "_get_incremental_prog_part", autospec=True) as mock_get_inc_part,
        ):
            # mocked return value of _remove_new_incremental_expressions
            mock_remove.side_effect = [[("function", ['"count"']), ("unary", ['"()"'])]]
            # expected program parts
            expected_parts = [
                ("new_function", [String('"count"'), Number(0)]),
                ("new_unary", [String('"()"'), Number(0)]),
            ]
            # mock the program parts as return values of _get_incremental_prog_part
            mock_get_inc_part.side_effect = expected_parts.copy()
            # expected calls to _get_incremental_prog_part
            expected_get_inc_part_calls = [
                call("function", ['"count"'], 0),
                call("unary", ['"()"'], 0),
            ]

            # compute program parts for bound 0
            parts = app._compute_prog_parts(0)

            # check that correct parts are returned
            self.assertEqual(parts, expected_parts)
            # check call to _get_incremental_prog_part
            self.assertEqual(mock_get_inc_part.call_args_list, expected_get_inc_part_calls)

    def test_compute_prog_parts_at_non_zero(self) -> None:
        """
        Test compute program parts function for non-zero bound.
        """
        app = COOMMultiSolverApp([])

        with (
            patch.object(app, "_remove_new_incremental_expressions", autospec=True) as mock_remove,
            patch.object(app, "_get_prog_part", autospec=True) as mock_get_part,
        ):
            # mock value of _incremental_parts
            app._incremental_parts = {("unary", tuple())}
            # mocked return value of _remove_new_incremental_expressions
            mock_remove.side_effect = [[]]

            # mocked return value of _get_prog_part
            non_inc_parts = [
                ("new_set", (String('"root.color"'), String('"root.color[0]"'))),
                ("new_set", (String('"root.bags"'), String('"root.bags[1]"'))),
            ]
            mock_get_part.side_effect = non_inc_parts.copy()

            # mocked return value of _get_prog_part_of_incremental_set
            inc_parts = [("incremental_unary", (Number(1),))]

            # initial value of _new_processed_facts
            new_facts = [
                'set("root.bags","root.bags[1]").',
                'set("root.color","root.color[0]").',
            ]
            app._new_processed_facts = set(new_facts)

            # compute all program parts for bound 1
            parts = app._compute_prog_parts(1)

            # check if the correct parts are returned
            self.assertEqual(
                parts, non_inc_parts + inc_parts, f"parts is {parts}, expected {non_inc_parts + inc_parts}"
            )
            # check calls to _get_prog_part
            self.assertCountEqual(mock_get_part.call_args_list, [call(x, 1) for x in new_facts])

    def test_main_control_calls(self) -> None:
        """
        Test calls to the control object in multi application main function.
        """
        app = COOMMultiSolverApp([], initial_bound=2)

        control = self._get_mock_control([False, True])

        with (
            redirect_stdout(None),
            patch.object(app, "_preprocess_new_bound", autospec=True),
            patch.object(app, "_compute_prog_parts", autospec=True),
            patch.object(app, "_find_minimal_bound", autospec=True),
        ):
            app.main(control, [])

            expected_calls = [
                call.load(ANY),
                call.load(ANY),
                # check that base program is grounded first
                call.add("base", [], ANY),
                call.ground([("base", [])]),
                # ground each bound in 0..2 and set the active external
                call.ground(ANY),
                call.assign_external(Function("active", [Number(0)], True), True),
                call.ground(ANY),
                call.assign_external(Function("active", [Number(1)], True), True),
                call.ground(ANY),
                call.assign_external(Function("active", [Number(2)], True), True),
                # check that max bound external is set correctly
                call.assign_external(Function("max_bound", [Number(2)], True), True),
                call.solve(),
                # grounding and assigning active for next bound
                call.ground(ANY),
                call.assign_external(Function("active", [Number(3)], True), True),
                # releasing previous max bound and assigning new max bound
                call.assign_external(Function("max_bound", [Number(3)], True), True),
                call.release_external(Function("max_bound", [Number(2)], True)),
                call.solve(),
            ]

            self.assertEqual(
                len(control.mock_calls),  # type: ignore[attr-defined]
                len(expected_calls),
                "number of calls to mock control does not match expected number of calls",
            )

            for i, (real, expected) in enumerate(zip(control.mock_calls, expected_calls)):  # type: ignore[attr-defined]
                self.assertEqual(real, expected, f"the {i}th call to mock control does not match the expected call")



import os
import sys

import nox

nox.options.sessions = "lint_pylint", "typecheck", "test"

EDITABLE_TESTS = True
PYTHON_VERSIONS = ["3.13"]
if "GITHUB_ACTIONS" in os.environ:
    PYTHON_VERSIONS = ["3.12", "3.13"]
    EDITABLE_TESTS = False


@nox.session
def doc(session):
    """
    Build the documentation.

    Accepts the following arguments:
    - serve: open documentation after build
    - further arguments are passed to mkbuild
    """

    options = session.posargs[:]
    open_doc = "serve" in options
    if open_doc:
        options.remove("serve")

    session.install("-e", ".[doc]")

    if open_doc:
        open_cmd = "xdg-open" if sys.platform == "linux" else "open"
        session.run(open_cmd, "http://localhost:8000/systems/coom-suite/")
        session.run("mkdocs", "serve", *options)
    else:
        session.run("mkdocs", "build", *options)


@nox.session
def dev(session):
    """
    Create a development environment in editable mode.

    Activate it by running `source .nox/dev/bin/activate`.
    """
    session.install("-e", ".[dev]")


@nox.session
def lint_pylint(session):
    """
    Run pylint.
    """
    session.install("-e", ".[lint_pylint,test]")
    session.run("pylint", "coomsuite", "tests")


@nox.session
def typecheck(session):
    """
    Typecheck the code using mypy.
    """
    session.install("-e", ".[typecheck,test]")
    session.run("mypy", "--strict", "-p", "coomsuite", "-p", "tests", "--ignore-missing-imports")


@nox.session(python=PYTHON_VERSIONS)
def test(session):
    """
    Run the tests.

    Accepts an additional arguments which are passed to the unittest module.
    This can for example be used to selectively run test cases.
    """

    args = [".[test]"]
    if EDITABLE_TESTS:
        args.insert(0, "-e")
    session.install(*args)
    if session.posargs:
        session.run("coverage", "run", "-m", "unittest", session.posargs[0], "-v")
    else:
        session.run("coverage", "run", "-m", "unittest", "discover", "-v")
        session.run("coverage", "report", "-m", "--fail-under=100")



"""
Contains a custom backend for the UI.
"""

import base64

from clingo import Control
from clinguin.server.application.backends.explanation_backend import ExplanationBackend
from clinguin.utils.annotations import overwrites

from coomsuite.utils import asp2coom, coom2asp


class CoomBackend(ExplanationBackend):  # type: ignore
    """
    Extends ExplanationBackend with functionality for the COOM UI.
    """

    @overwrites(ExplanationBackend)  # type: ignore
    def download(self, show_prg=None, file_name="current_solution.coom") -> None:  # type: ignore
        """
        Downloads the current model in COOM format. It must be selected first via :func:`~select` .

        Arguments:
            show_prg (_type_, optional): Program to filter output using show statements. Defaults to None.
            file_name (str, optional): The name of the file for the download. Defaults to "current_solution.coom".
        """
        if self._model is None:
            raise RuntimeError("Cant download when there is no model")
        show_prg = show_prg or ""
        prg = "\n".join([f"{s}." for s in self._model])
        ctl = Control()
        ctl.add("base", [], prg)
        try:
            ctl.add("base", [], show_prg.replace('"', ""))
        except RuntimeError as exc:
            raise RuntimeError("Show program can't be parsed. Make sure it is a valid clingo program.") from exc
        ctl.ground([("base", [])])
        with ctl.solve(yield_=True) as hnd:
            for m in hnd:
                output_symbols = list(m.symbols(shown=True))

        sorted_symbols = sorted(output_symbols)
        coom_solution = "\n".join([f"{asp2coom(s)}" for s in sorted_symbols])

        file_name = file_name.strip('"')
        with open(file_name, "w", encoding="UTF-8") as file:
            file.write(coom_solution)
        self._messages.append(
            (
                "Download successful",
                f"Information saved in file {file_name}.",
                "success",
            )
        )

    @overwrites(ExplanationBackend)  # type: ignore
    def upload_file(self) -> None:
        """
        Upload file using the context. The context should have the name of the file under `filename` and the
        file content in base64 under `filecontent`.
        """
        coom_solution = next((c.value for c in self._context if c.key == "filecontent"), None)

        if not coom_solution:
            raise RuntimeError("No content found in context")

        decoded_coom_solution = base64.b64decode(coom_solution).decode("utf-8")
        asp_facts = [coom2asp(l) for l in decoded_coom_solution.splitlines() if l.strip()]
        self.clear_assumptions()

        for f in asp_facts:
            self.add_assumption(f)



"""
Utilities.
"""

from importlib.resources import as_file, files
from typing import List

from antlr4 import CommonTokenStream, InputStream
from clingo import Symbol

from coomsuite.utils.parse_coom import ASPModelVisitor, ASPUserInputVisitor

from .coom_grammar.model.ModelLexer import ModelLexer
from .coom_grammar.model.ModelParser import ModelParser
from .coom_grammar.user.UserInputLexer import UserInputLexer
from .coom_grammar.user.UserInputParser import UserInputParser

# mypy: allow-untyped-calls


def get_encoding(file_name: str) -> str:  # nocoverage
    """Gets the path to a given ASP encoding in the encodings folder

    Args:
        file_name (str): The name of the ASP encoding with the extension

    Returns:
        str: The string for the path to the ASP encoding
    """
    with as_file(files("coomsuite") / "encodings" / file_name) as file:
        return str(file)


def run_antlr4_visitor(coom_input_stream: InputStream, grammar: str) -> List[str]:
    """Runs the ANTLR4 Visitor.

    Args:
        coom_input_stream (antlr4.InputStream): The input COOM encoding as a file stream

    Returns:
        List[str]: The converted ASP instance

    """
    if grammar == "model":
        lexer = ModelLexer(coom_input_stream)
        stream = CommonTokenStream(lexer)
        parser = ModelParser(stream)
        tree = parser.root()
        visitor = ASPModelVisitor()
        visitor.visitRoot(tree)
    elif grammar == "user":
        lexer = UserInputLexer(coom_input_stream)
        stream = CommonTokenStream(lexer)
        parser = UserInputParser(stream)
        tree = parser.user_input()
        visitor = ASPUserInputVisitor()
        visitor.visitUser_input(tree)
    return visitor.output_asp


def asp2coom(s: Symbol) -> str:
    """
    Formats ASP output symbols to a more readable COOM format.
    """
    if s.name == "include":
        return s.arguments[0].string.removeprefix("root.")
    if s.name == "value":
        path = s.arguments[0].string.removeprefix("root.")
        value = s.arguments[1]
        return f"{path} = {value}"
    raise ValueError(f"Unrecognized predicate: {s.name}")


def coom2asp(c: str) -> str:
    """
    Converts COOM facts to ASP facts.
    """
    if "=" in c:
        path, value = c.split("=")
        return f'value("root.{path.strip()}",{value.strip()})'
    path = c.strip()
    return f'include("root.{path}")'



"""
Test cases for bound functionalities.
"""

from contextlib import redirect_stdout
from unittest import TestCase
from unittest.mock import call, patch

from coomsuite.bounds import get_bound_iter, next_bound_converge
from coomsuite.bounds.solver import BoundSolver


class TestBound(TestCase):
    """
    Test cases for bound iterator and convergence functions.
    """

    def test_bound_iter_linear(self) -> None:
        """
        Test the linear bound iterator.
        """
        bound_iter = get_bound_iter(algorithm="linear", start=0, step=1)
        self.assertEqual(next(bound_iter), 0)
        self.assertEqual(next(bound_iter), 1)
        self.assertEqual(next(bound_iter), 2)
        self.assertEqual(next(bound_iter), 3)

        bound_iter = get_bound_iter(algorithm="linear", start=5, step=2)
        self.assertEqual(next(bound_iter), 5)
        self.assertEqual(next(bound_iter), 7)
        self.assertEqual(next(bound_iter), 9)

    def test_bound_iter_exponential(self) -> None:
        """
        Test the exponential bound iterator.
        """
        bound_iter = get_bound_iter(algorithm="exponential", start=0, step=2)
        self.assertEqual(next(bound_iter), 0)
        self.assertEqual(next(bound_iter), 1)
        self.assertEqual(next(bound_iter), 2)
        self.assertEqual(next(bound_iter), 4)
        self.assertEqual(next(bound_iter), 8)

        bound_iter = get_bound_iter(algorithm="exponential", start=3, step=2)
        self.assertEqual(next(bound_iter), 3)
        self.assertEqual(next(bound_iter), 4)
        self.assertEqual(next(bound_iter), 8)

        bound_iter = get_bound_iter(algorithm="exponential", start=4, step=3)
        self.assertEqual(next(bound_iter), 4)
        self.assertEqual(next(bound_iter), 9)
        self.assertEqual(next(bound_iter), 27)

    def test_bound_iter_unknown(self) -> None:
        """
        Test invalid input for algorithm of bound iterator.
        """
        self.assertRaises(ValueError, get_bound_iter, "test", 0, 1)

    def test_converge(self) -> None:
        """
        Test computation of bound for convergence step.
        """
        self.assertEqual(next_bound_converge(3, 4), None)
        self.assertEqual(next_bound_converge(8, 16), 12)
        self.assertEqual(next_bound_converge(-1, 4), 1)
        self.assertEqual(next_bound_converge(3, 8), 5)

    def test_singleshot_converge(self) -> None:
        """
        Test converge function used in single shot solving.
        """
        solver = BoundSolver([], solver="clingo", clingo_args=[], output_format="asp")

        with redirect_stdout(None):
            for init_unsat, init_sat, solve_returns, max_bound in [
                (4, 8, [10, 10], 5),
                (4, 8, [10, 20], 6),
                (4, 8, [20, 10], 7),
                (4, 8, [20, 20], 8),
                (4, 8, [65], None),
            ]:
                # returns = [10 if x else 20 for x in solve_returns]
                with (
                    patch.object(solver, "_solve", autospec=True, side_effect=solve_returns),
                    redirect_stdout(None),
                ):
                    ret = solver._converge(init_unsat, init_sat)  # pylint: disable=protected-access
                    self.assertEqual(
                        ret,
                        max_bound,
                        (
                            f"failed with init_unsat={init_unsat}, init_sat={init_sat}, "
                            f"solve_returns={solve_returns}, max_bound={max_bound}"
                        ),
                    )
        # Test exception for invalid error code
        with self.assertRaises(KeyError, msg="failed with exit code=99"):
            with patch.object(solver, "_solve", autospec=True, side_effect=[99]), redirect_stdout(None):
                ret = solver._converge(4, 8)  # pylint: disable=protected-access

    def test_get_bounds_singleshot(self) -> None:
        """
        Test get bounds function for singleshot solving.
        """
        solver = BoundSolver([], solver="clingo", clingo_args=[], output_format="asp")

        for algorithm, initial_bound, step, solve_returns, converge_return, expected_converge_call in [
            ("linear", 0, None, [10], 0, call(-1, 0)),
            ("linear", 2, 2, [20, 10], 3, call(2, 4)),
            ("linear", 1, 1, [65], None, None),
            ("exponential", 3, 2, [10], 1, call(-1, 3)),
            ("exponential", 3, 2, [20, 10], 4, call(3, 4)),
            ("exponential", 1, 2, [65], None, None),
        ]:
            with (
                patch.object(solver, "_solve", autospec=True, side_effect=solve_returns) as mock_solve,
                patch.object(solver, "_converge", autospec=True, side_effect=[converge_return]) as mock_converge,
            ):
                # mock_solve.side_effect = solve_returns  # [10 if x else 20 for x in solve_returns]
                # mock_converge.side_effect = [converge_return]

                with redirect_stdout(None):
                    minimal_bound = solver.get_bounds(algorithm, initial_bound, step, use_multishot=False)

                fail_msg = (
                    f"failed with algorithm={algorithm}, initial_bound={initial_bound}, solve_returns={solve_returns}, "
                    f"converge_return={converge_return}, expected_converge_call={expected_converge_call}"
                )

                self.assertEqual(minimal_bound, converge_return, fail_msg)
                self.assertEqual(mock_converge.call_args, expected_converge_call, fail_msg)

        # Test exception for invalid exit code
        for algorithm in ["linear", "exponential"]:
            with self.assertRaises(KeyError, msg=f"failed with algorithm={algorithm} and exit code=99"):

                with (
                    patch.object(solver, "_solve", autospec=True) as mock_solve,
                    patch.object(solver, "_converge", autospec=True) as mock_converge,
                ):
                    mock_solve.side_effect = [99]

                    with redirect_stdout(None):
                        solver.get_bounds(algorithm, 0, use_multishot=False)



"""
Clingo application class for solving COOM configuration problems with multi-shot solving.
"""

from typing import Any, Dict, Iterator, List, Optional, Sequence, Set, Tuple, TypeAlias

from clingo import Control
from clingo.symbol import Function, Number, Symbol, parse_term

from coomsuite.application import COOMSolverApp
from coomsuite.preprocess import preprocess
from coomsuite.utils import get_encoding

from . import get_bound_iter, next_bound_converge

ProgPart: TypeAlias = Tuple[str, Tuple[Symbol, ...]]


def print_prog_parts(parts: List[ProgPart], name_filter: List[str] | None = None) -> None:  # nocoverage
    """
    Print a list of program parts, optionally include a filter (useful for debugging).

    The name_filter is a list of program part names. Any part names not included in the filter are not printed.
    Note that any prefixes of program parts (such as new_, update_, incremental_) as well as suffixes (_r, _l)
    do not need to be included in the name_filter.
    """
    for part in parts:
        name = part[0]

        x = name
        for prefix in ["update_", "new_", "incremental_"]:
            x = x.removeprefix(prefix)
        for suffix in ["_r", "_l"]:
            x = x.removesuffix(suffix)
        if name_filter and x not in name_filter:
            continue

        part_str = name
        part_str += "("
        for i, arg in enumerate(part[1]):
            part_str += str(arg)
            if i < len(part[1]) - 1:
                part_str += ","
        part_str += ")"
        print(part_str)


def _get_fact_name_and_args(fact: str) -> ProgPart:
    """
    Convert a fact given as a string to its name and arguments

    Args:
        fact (str): The fact in string representation (with trailing '.')

    Returns:
        ProgPart: A tuple of the name of the fact and the tuple of its arguments
    """
    x = parse_term(fact[:-1])
    return (x.name, tuple(x.arguments))


class COOMMultiSolverApp(COOMSolverApp):  # pylint: disable=too-many-instance-attributes
    """
    Application class for multi-shot solving extending the standard COOM application class
    """

    def __init__(
        self,
        serialized_facts: List[str],
        step: Optional[int] = None,
        algorithm: str = "linear",
        initial_bound: int = 0,
        log_level: str = "",
        options: Optional[Dict[str, Any]] = None,
        istest: bool = False,
    ):  # pylint: disable=too-many-arguments,too-many-positional-arguments
        super().__init__(log_level, options, istest)

        if self._options["solver"] == "flingo":
            raise ValueError("multishot solving is currently not supported for the flingo solver")

        self._serialized_facts: List[str] = serialized_facts
        """The instance to solve given as serialized facts"""

        if step is None:
            step = 1 if algorithm == "linear" else 2
        self._bound_iter: Iterator[int] = get_bound_iter(algorithm, initial_bound, step)
        """Iterator determining the bounds"""
        self.current_max_bound: int = next(self._bound_iter)
        """The current max bound"""

        self._prev_bound: Optional[int] = None
        """The previous max bound"""

        self._new_processed_facts: Set[str] = set()
        """Processed facts added by current preprocessing step"""
        self._processed_facts: Set[str] = set()
        """Processed facts from all previous preprocessing steps"""

        self._incremental_parts: Set[ProgPart] = set()
        """The set of all incremental program parts"""
        self._incremental_expressions: Set[str] = set()
        """The set of all incremental expressions (represented by their name)"""
        self._is_initialized: Set[str] = set()
        """Keeps track of all incremental expressions that are already initialized"""

    def _update_bound(self) -> None:
        """
        Update the current maximum bound
        """
        self._prev_bound = self.current_max_bound
        self.current_max_bound = next(self._bound_iter)

    def _preprocess_new_bound(self, bound: int) -> None:
        """
        Preprocess the serialized facts for the given bound and update the fact data structures

        Args:
            bound (int): the value of max_bound to preprocess
        """
        # update facts that were already processed
        self._processed_facts.update(self._new_processed_facts)

        # preprocess with bound
        facts = set(preprocess(self._serialized_facts, max_bound=bound, discrete=True, multishot=True))

        # split facts into incremental and non-incremental facts
        incremental_facts = set()
        non_incremental_facts = set()
        for x in facts:
            if x.startswith(("inc_set", "incremental")):
                incremental_facts.add(x)
            else:
                non_incremental_facts.add(x)

        # update incremental data with results from preprocessing
        self._update_incremental_data(incremental_facts)

        # filter out facts that were previously processed
        self._new_processed_facts = non_incremental_facts - self._processed_facts

    def _remove_new_incremental_expressions(self) -> List[ProgPart]:
        """
        Remove all facts from new_processed_facts that are incremental expressions

        Program parts for incremental expressions are added by iterating over self._incremental_parts.
        This is done because the facts for incremental expressions are only added once,
        but then have to be updated for every new bound. To ensure that any new incremental
        facts are also added in the correct way they are removed from the _new_processed_facts.

        Returns:
            List[ProgPart]: list of incremental expressions represented as tuples (name, args)
        """
        inc_expressions = []
        for fact in self._new_processed_facts:
            name, args = _get_fact_name_and_args(fact)

            # check if the fact is an incremental constraint
            is_incremental_constraint = (
                name == "constraint"
                and args[1].string == "boolean"
                and args[0].arguments[1].string in self._incremental_expressions
            )
            # check if the fact is an incremental expression
            is_incremental_expression = (
                name in ["function", "binary", "unary", "minimize", "maximize"]
                and args[0].string in self._incremental_expressions
            )

            if is_incremental_constraint or is_incremental_expression:
                inc_expressions.append((name, args))
                # add the fact to the processed_facts
                self._processed_facts.add(fact)

        # filter new_processed_facts to remove the incremental expressions
        self._new_processed_facts = self._new_processed_facts - self._processed_facts

        return inc_expressions

    def _get_incremental_prog_part(self, exp_type: str, args: Tuple[Symbol, ...], bound: int) -> ProgPart:
        """
        Get the incremental program part for an expression

        Args:
            exp_type (str): string representing the type of the expression
            args (Tuple[Symbol, ...]): the arguments of the expression
            bounds (int): the current bound to use as parameter of the program part

        Returns:
            ProgPart: the program part containing the rules to update the incremental expression
        """
        # determine the name and arguments of the program part
        part_name = ""
        # to the arguments we just need to add the current max_bound
        args += (Number(bound),)
        # the name of the program part depends on the type of the expression
        match exp_type:
            case "unary" | "constraint" | "minimize" | "maximize":
                part_name = "incremental_" + exp_type
            case "function":
                # determine the name of the function
                name = args[0].string
                # to determine the part name we need to check if the function is initialized
                prefix = "update_" if name in self._is_initialized else "new_"
                part_name = prefix + "incremental_function"
                # mark the function as initialized
                self._is_initialized.add(name)
            case "binary":
                # for binaries we need to check which of the subexpressions are incremental themselves
                part_name = "incremental_binary"
                lhs = args[1].string
                rhs = args[3].string

                lhs_inc = lhs in self._incremental_expressions
                rhs_inc = rhs in self._incremental_expressions
                if lhs_inc and not rhs_inc:
                    part_name += "_l"
                elif not lhs_inc and rhs_inc:
                    part_name += "_r"
                elif not lhs_inc and not rhs_inc:
                    # one subexpression has to be incremental as otherwise the binary would not be incremental
                    raise ValueError(f"Incremental binary but no subexpression is incremental: {args}")  # nocoverage
                # if both lhs and rhs are incremental no suffix is added
            case _:
                raise ValueError(f"unknown type of incremental expression: {exp_type}")

        return (part_name, args)

    def _get_prog_part(self, fact: str, bound: int) -> ProgPart:
        """
        Get the program part belonging to a fact

        Args:
            fact (str): The fact representet as a string (with trailing '.')
            bound (int): the current bound to use as parameter of the program part

        Returns:
            ProgPart: the program part containing the rules for the fact
        """
        # convert fact to name and arguments
        name, args = _get_fact_name_and_args(fact)

        # check if the fact corresponds to a valid program part
        if name not in [
            "parent",
            "index",
            "set",
            "type",
            "constraint",
            "column",
            "unary",
            "binary",
            "function",
            "allow",
            "number",
            "constant",
            "minimize",
            "maximize",
        ]:
            raise ValueError(f"unknown new fact (no corresponding program part exists): {fact}")  # nocoverage

        # some program parts need the current bound as an additional argument
        needs_bound = False
        if name in ["type", "constraint", "column"]:
            needs_bound = True

        if needs_bound:
            args += (Number(bound),)

        # determine the corresponding program part
        program_part = (f"new_{name}", args)

        return program_part

    def _update_incremental_data(self, incremental_facts: Set[str]) -> None:
        """
        Update internal data structures keeping track of the incremental sets and their expressions

        Args:
            incremental_facts (Set[str]): a set of incremental facts to add to the internal incremental data structures
        """
        # incremental_facts contains the predicate:
        # incremental(T,N,Args) indicating an expression of type T with name N is incremental
        #                       Args are the arguments of the expression
        inc_expressions = {parse_term(x[:-1]) for x in incremental_facts if x.startswith("incremental")}

        # add incremental expressions
        for exp in inc_expressions:
            exp_type = exp.arguments[0].string
            exp_name = exp.arguments[1].string
            exp_args = tuple(exp.arguments[2].arguments)

            # first, add the expressions to the set of incremental parts
            # for incremental expressions of type path this is not done, as paths are only auxiliary
            # and only used for the set of all incremental expressions (below)
            if exp_type != "path":
                self._incremental_parts.add((exp_type, exp_args))

            # second, add its name to the set of all incremental expressions
            self._incremental_expressions.add(exp_name)

    def _find_minimal_bound(self, control: Control) -> None:
        """
        Find the minimal bound for an instance once a satisfiable bound was found

        This assumes that for _prev_bound the instance was UNSAT, and for max_bound the instance is SAT.
        The minimal bound is found when _prev_bound + 1 is equal to max_bound.

        After returning max_bound is the minimal bound for which the program is satisfiable.

        Args:
            control (Control): the control object (already containing all program parts grounded)
        """
        # unsat_bound and sat_bound give the range of the optimal bound
        unsat_bound = -1 if self._prev_bound is None else self._prev_bound
        sat_bound = self.current_max_bound
        # last_bound stores the bound from the last solve call
        last_bound = self.current_max_bound

        while True:
            # compute next bound to check
            current_bound = next_bound_converge(unsat_bound, sat_bound)

            # check if we found the optimal bound
            if current_bound is None:
                print("\nOptimal bound found")
                self.current_max_bound = sat_bound
                break

            # if we do not have the optimal bound yet we do another solve
            print("\nOptimal bound not yet found")
            print(f"Solving with bound = {current_bound}\n")

            # set active externals
            if current_bound < last_bound:
                # if the bound we want to solve at is smaller than the last bound,
                # the active externals have to be set to false
                for i in range(current_bound + 1, last_bound + 1):
                    control.assign_external(Function("active", [Number(i)]), False)
            else:
                # otherwise they have to be set to true
                for i in range(last_bound + 1, current_bound + 1):
                    control.assign_external(Function("active", [Number(i)]), True)

            # set max_bound externals
            control.release_external(Function("max_bound", [Number(last_bound)]))
            control.assign_external(Function("max_bound", [Number(current_bound)]), True)

            ret = control.solve()
            # update bound of last solve call
            last_bound = current_bound
            # update sat/unsat bound depending on result of solve
            if ret.satisfiable:
                sat_bound = current_bound
            else:
                unsat_bound = current_bound

    def _compute_prog_parts(self, bound: int) -> List[ProgPart]:
        """
        Compute all program parts for the current new processed facts

        Args:
            bound (int): the current bound to compute program parts for

        Returns:
            List[ProgPart]: the list of all program parts needed for bound
        """
        parts = []

        # remove all expressions that were detected as incremental from the processed facts
        # (necessary to avoid adding them via the normal new_* program parts,
        # instead we add them using specific incremental program parts later)
        incremental_expressions = self._remove_new_incremental_expressions()

        if bound == 0:
            # add the incremental program parts corresponding to each incremental expression
            for name, args in incremental_expressions:
                parts.append(self._get_incremental_prog_part(name, args, bound))
        else:
            # collect program parts for all the new facts
            for fact in self._new_processed_facts:
                parts.append(self._get_prog_part(fact, bound))

            # add the program parts belonging to every incremental set
            for part in self._incremental_parts:
                parts.append(self._get_incremental_prog_part(part[0], part[1], bound))

        return parts

    def main(self, control: Control, files: Sequence[str]) -> None:
        """
        Main function of the multishot application class

        After returning the attribute max_bound is the minimal bound for which the instance is satisfiable.

        Args:
            control (Control): the clingo control object
        """
        control.load(get_encoding("encoding-base-clingo-multi.lp"))
        control.load(get_encoding("show-clingo.lp"))

        while True:
            print(f"\nNew max bound is = {self.current_max_bound} (previous was {self._prev_bound})\n")

            # grounding and assigning active externals
            # in steps of 1 to enable deactivation of specific bounds later (when converging to minimal bound)
            for bound in range(0 if self._prev_bound is None else self._prev_bound + 1, self.current_max_bound + 1):
                print(f"Grounding with bound = {bound}")

                # preprocessing
                self._preprocess_new_bound(bound)

                # collect program parts
                parts = self._compute_prog_parts(bound)

                if bound == 0:
                    # ground base with the remaining preprocessed facts added
                    # (needs to be grounded before other program parts below)
                    control.add("base", [], "".join(self._new_processed_facts))
                    control.ground([("base", [])])

                # ground
                control.ground(parts)

                # update active external
                control.assign_external(Function("active", [Number(bound)]), True)

            # update max bound external
            control.assign_external(Function("max_bound", [Number(self.current_max_bound)]), True)
            if self._prev_bound is not None:
                control.release_external(Function("max_bound", [Number(self._prev_bound)]))

            # solve
            print(f"\nSolving with bound = {self.current_max_bound}\n")
            ret = control.solve()
            if ret.satisfiable:
                self._find_minimal_bound(control)
                break

            self._update_bound()
